import inspect

from biomatools.utils.components import BioMALocationTable, BioMATable, BioMAWeatherTable
from biomatools.utils.formats import is_bioma_location, is_bioma_table, is_bioma_weather
from biomatools.utils.naming import BIOMA_LOCATION_NAME, BIOMA_WEATHER_NAME


class ModelResult():

    def __init__(self, Tables:list):
        self.data = {}
        for t in Tables:
            if is_bioma_table(t):
                if  is_bioma_location(t):
                    self.data[BIOMA_LOCATION_NAME] = BioMALocationTable.from_dict(t)
                elif is_bioma_weather(t):
                    self.data[BIOMA_WEATHER_NAME] = BioMAWeatherTable.from_dict(t)
                else:
                    self.data[t.get("Name")] = BioMATable.from_dict(t)
    
    def list_tables(self):
        return self.data.keys()
    
    def get_data_table(self, key:str):
        return self.data.get(key)


    @classmethod
    def from_dict(cls, env):
        '''This is mostly used to build the object from the output returned by Web Services'''
        return cls(**{
                k: v for k, v in env.items() 
                if k in inspect.signature(cls).parameters
            })
       



BIOMA_WEATHER_COLUMNS =  [ "AtmosphericPressure", "CO2concentration", "WaterTemperatureMinimum", "WaterTemperatureMaximum", "IsEvapotrPresent", "IsVpdPresent", "IsHumidityPresent", "AirRelativeHumidityMinimum", "Windspeed", "VapourPressureDeficit", "SnowDepth", "AirTemperatureMinimum", "AirTemperatureMaximum", "SoilEvaporation", "ReferenceEvapotranspiration", "SurfaceOzoneConcentration", "GlobalSolarRadiation", "IsCO2Present", "Date", "Average_temperature", "Precipitation", "AirRelativeHumidityMaximum", "LeafWetnessDuration", "Grid_no" ]  
BIOMA_LOCATION_COLUMNS = [ "Altitude", "Grid_no", "Latitude", "Longitude", "Distance_to_coast" ]
BIOMA_WEATHER_NAME = 'weather'
BIOMA_LOCATION_NAME = 'location'

BIOMA_SIMPLE_TYPES = ['System.DateTime', 'System.String']


import logging
import numpy as np
import pandas as pd

from biomatools.utils.abs import (
    AbstractBioMAClient,
    AbstractBioMAComponent,
    AbstractBioMAObject,
)
from biomatools.utils.naming import (
    BIOMA_LOCATION_NAME,
    BIOMA_WEATHER_COLUMNS,
    BIOMA_WEATHER_NAME,
)


def bioma_table_to_dataframe(table):
    """converts a BioMA data table into a pandas DataFrame object indexed by datetime"""
    # telling pandas which are the fields containing column names and actual rows
    data_pd = pd.DataFrame(table["Rows"], columns=table["ColumnNames"])
    # parsing dates...
    data_pd["Date"] = pd.to_datetime(data_pd["Date"])
    # ... and setting them as rows index
    data_pd = data_pd.set_index("Date")
    return data_pd


def weather_table_to_dataframe(weather_table):
    """converts a BioMA weather data table into a pandas DataFrame object indexed by datetime"""
    # telling pandas which are the fields containing column names and actual rows
    weather_data_pd = bioma_table_to_dataframe(weather_table)
    # telling pandas the correct column type
    weather_data_pd["Average_temperature"] = weather_data_pd.Average_temperature.astype(
        "float"
    )
    return weather_data_pd


def prepare_df(df: pd.DataFrame, col_list, renaming=None) -> dict:
    """As BioMA tables are expected to have a fixed number of columns
    with fixed labels, we'd better format dataframes before using them
    to build Weather and/or Location objects.

    Parameters
    ----------
    df: pd.DataFrame
        the dataframe you wish to prepare for BioMA ingestion

    col_list: list
        the list of column labels the BioMA object should have

    renaming: dict
        a dictionary that desribes a renaming operation, in the form
        {"old name A": "new name A", "old name B": "new name B"}.
        Defailt: None
    """
    if renaming:
        df = df.rename(columns=renaming)
    # first let's remove columns that should not be there
    df = df.drop([c for c in df.columns if c not in col_list], axis=1)
    # than add NA columns
    for c in col_list:
        if c not in df.columns:
            df[c] = np.nan
    return df


def is_bioma_table(data: dict, expected_columns=None) -> bool:
    """Tells if a dictionary is an acceptable BioMA data table"""
    # the following is not true anymore
    # if not isinstance(data.get("NotNullFields"), list):
    #    return False
    if not isinstance(data.get("ColumnNames"), list):
        return False
    if isinstance(data.get("NotNullFields"), list):
        if len(data.get("NotNullFields")) > len(data.get("ColumnNames")):
            # how can not null fields be more than the total number of fields?
            return False
    else:
        if expected_columns is not None:
            if any([i not in expected_columns for i in data.get("ColumnNames")]):
                # this means we fire a "False" if we find some column name we would not expect
                return False
    if not isinstance(data.get("Rows"), list):
        return False
    # we did not find anything compromising here
    return True


def is_bioma_weather(data: dict) -> bool:
    if data.get("Name") == BIOMA_WEATHER_NAME:
        return is_bioma_table(data, expected_columns=BIOMA_WEATHER_COLUMNS)
    else:
        return False


def is_bioma_location(data: dict) -> bool:
    if data.get("Name") == BIOMA_LOCATION_NAME:
        return is_bioma_table(data, expected_columns=BIOMA_WEATHER_COLUMNS)
    else:
        return False


def json_serialize_component(c):
    if isinstance(c, AbstractBioMAObject):
        return c.json_serialize()
    else:
        return vars(c)


def check_bioma_type(
    expected_obj: AbstractBioMAComponent,
    found_obj: AbstractBioMAComponent,
    client: AbstractBioMAClient,
):
    """Checks if a BioMACOnfiguration belongs to the correct type"""
    if expected_obj.BioMAtype == found_obj.BioMAtype:
        return True
    if len(expected_obj.subtypes) == 0:
        try:
            expected_obj.load_subtypes(client)
        except:
            logging.info(f"Object {expected_obj.BioMAtype} has no subtypes")
    return found_obj.BioMAtype in [a.BioMAtype for a in expected_obj.subtypes]



from biomatools.utils import formats

false= False
weather_table_example = {
			"ColumnNames": [
				"AtmosphericPressure",
				"CO2concentration",
				"WaterTemperatureMinimum",
				"WaterTemperatureMaximum",
				"LeafWetnessDuration",
				"IsEvapotrPresent",
				"IsVpdPresent",
				"IsHumidityPresent",
				"AirRelativeHumidityMinimum",
				"AirRelativeHumidityMaximum",
				"Windspeed",
				"VapourPressureDeficit",
				"SnowDepth",
				"Precipitation",
				"Average_temperature",
				"AirTemperatureMinimum",
				"AirTemperatureMaximum",
				"Grid_no",
				"SoilEvaporation",
				"ReferenceEvapotranspiration",
				"Date",
				"SurfaceOzoneConcentration",
				"GlobalSolarRadiation",
				"IsCO2Present"
			],
			"Name": "weather",
			"Rows": [
				[
					"NaN",
					"NaN",
					"NaN",
					"NaN",
					0,
					false,
					false,
					false,
					"NaN",
					70.6,
					"NaN",
					"NaN",
					"NaN",
					0.0,
					6.4,
					"NaN",
					6.4,
					"50",
					"NaN",
					"NaN",
					"2021-03-02T00:00:00Z",
					"NaN",
					"NaN",
					false
				],
				[
					"NaN",
					"NaN",
					"NaN",
					"NaN",
					0,
					false,
					false,
					false,
					"NaN",
					70.5,
					"NaN",
					"NaN",
					"NaN",
					0.0,
					6.3,
					"NaN",
					6.3,
					"50",
					"NaN",
					"NaN",
					"2021-03-02T01:00:00Z",
					"NaN",
					"NaN",
					false
				]
            ]
    }

location_table_example = {
			"ColumnNames": [
				"Grid_no",
				"Altitude",
				"Latitude",
				"Longitude",
				"Distance_to_coast"
			],
			"Name": "locations",
			"Rows": [
				[
					"50",
					-1000,
					"NaN",
					"NaN",
					"NaN"
				]
			]
		}

def test_bioma_table():
    assert formats.is_bioma_table(location_table_example)
'''
def test_bioma_location_table():
    assert formats.is_bioma_location(location_table_example)

def test_bioma_weather_table():
    assert formats.is_bioma_weather(weather_table_example)
'''


from abc import ABC, abstractmethod
import json


class AbstractBioMAClient(ABC):
    @abstractmethod
    def get_initial_configuration(self):
        pass

    @abstractmethod
    def get_object_configuration(self, assembly_qualified_name: str):
        pass

    @abstractmethod
    def get_object_subtypes(self, assembly_qualified_name: str):
        pass

    @abstractmethod
    def list_models(self):
        pass

    @abstractmethod
    def run(self, idx):
        pass

    @abstractmethod
    def run_all(self):
        pass


class AbstractBioMAObject(ABC):

    def toJSON(self) -> str:
        """Serializes item into a JSON STRING"""
        return json.dumps(self.json_serialize())

    def to_dict(self) -> dict:
        """Serializes item into a Dictionary"""
        return vars(self)

    @abstractmethod
    def json_serialize(self):
        """turns the item into something that is JSON serializable"""
        pass


class AbstractBioMATable(AbstractBioMAObject):

    @abstractmethod
    def to_df(self):
        pass


class AbstractBioMAComponent(AbstractBioMAObject):
    @abstractmethod
    def load_configurations(self, client: AbstractBioMAClient):
        pass

    @abstractmethod
    def load_subtypes(self, client: AbstractBioMAClient, recursive: bool = False):
        pass

    @abstractmethod
    def set_param(self, parname: str, parvalue=None):
        pass

    @abstractmethod
    def describe(self):
        pass



import logging
from biomatools.utils.components import BioMALocationTable, BioMAWeatherTable, InvalidLocationDictionaryError, InvalidWeatherDictionaryError
import pandas as pd

from biomatools.utils.naming import BIOMA_LOCATION_COLUMNS, BIOMA_WEATHER_COLUMNS

class ModelRun():

    def __init__(self, weather:BioMAWeatherTable=None, locations:BioMALocationTable=None, params=None, configuration = None):
        self.weather = weather
        self.locations = locations
        self.params = params
        self.configuration = configuration


    def set_weather_data(self, data = None, *args, **kwargs):
        f'''sets the weather data table no matter what you throw in it.
        
        Parameters
        ----------
        data: dict | pd.DataFrame default:None
            the weather data you want to include in the model run.
            Use this argument if you already have the all of it into a data structure
            and that data structure is a dictionary already formatted as a BioMA table
            or a pandas DataFrame. If you don't have such a data structure, just pass
            data series as individual arguments
        renaming: dict default:None
            a dictionary of mappings between your data format names and BioMA tables
            variables names. You need this only if you are passing a pandas DataFrame
            and column names are not set properly.
        *args and **kwargs
            if what you have is a bunch of lists, pass them individually as *named arguments*
            valid variable names are {BIOMA_WEATHER_COLUMNS}
        
        '''
        if isinstance(data, dict):
            try:
                self.weather = BioMAWeatherTable.from_dict(data)
            except InvalidWeatherDictionaryError as e:
                # wrong dictionary structure and/or naming
                logging.error('Invalid dictionary structure and/or naming. Check the documentation.')
                logging.info(help(BioMAWeatherTable))
                raise e
        elif isinstance(data, pd.DataFrame):
            # this fails rather silently: if the dataframe is not parsed correctly, it will
            # just create an empty one
            self.weather = BioMAWeatherTable.from_df(data, *args, **kwargs)
       
        elif data is None:
            self.weather = BioMAWeatherTable.from_series(*args, **kwargs)

    def set_locations(self, data, *args, **kwargs):
        f'''sets the weather data table no matter what you throw in it.
        
        Parameters
        ----------
        data: dict | pd.DataFrame default:None
            the location data you want to include in the model run.
            Use this argument if you already have the all of it into a data structure
            and that data structure is a dictionary already formatted as a BioMA table
            or a pandas DataFrame. If you don't have such a data structure, just pass
            data series as individual arguments
        renaming: dict default:None
            a dictionary of mappings between your data format names and BioMA tables
            variables names. You need this only if you are passing a pandas DataFrame
            and column names are not set properly.
        *args and **kwargs
            if what you have is a bunch of lists, pass them individually as *named arguments*
            valid variable names are {BIOMA_LOCATION_COLUMNS}
        
        '''
        if isinstance(data, dict):
            try:
                self.locations = BioMALocationTable.from_dict(data)
            except InvalidLocationDictionaryError as e:
                # wrong dictionary structure and/or naming
                logging.error('Invalid dictionary structure and/or naming. Check the documentation.')
                logging.info(help(BioMALocationTable))
                raise e
        elif isinstance(data, pd.DataFrame):
            # this fails rather silently: if the dataframe is not parsed correctly, it will
            # just create an empty one
            self.locations = BioMALocationTable.from_df(data, *args, **kwargs)
       
        elif data is None:
            self.locations= BioMALocationTable.from_series(*args, **kwargs)

    def set_configuration(self,):
        pass

    def set_parameters(self,):
        pass

    def build_request(self):
        out = {'tables': [self.locations.toJSON(), 
                          self.weather.toJSON()]}
        return out


class SimplifiedModelRun(ModelRun):
    
    def __init__(self, weather):
        super().__init__(weather, None, None, None)


import logging
import os

from biomatools.bioma_model import BioMAConfiguration
from biomatools.model_result import ModelResult
from biomatools.utils.abs import AbstractBioMAComponent
from biomatools.utils.connection import call_restful_api

class ModelClient():
    '''The client for a BioMA cloud model'''

    def __init__(self, url, key=None, env_variable_name = 'BIOMA_API_KEY'):
        self.url =url
        self.runs = {} # here we store model run requests
        self.outputs = {} # here we store results
        if key is None:
            env_key = os.environ.get(env_variable_name)
            if env_key is None:
                logging.warning('No API key found in environment')
                self.key = None
            else:
                logging.info(f'Using variable {env_variable_name} from local environment as API key')
                self.key = env_key
        else:
            self.key = key
        self.initial_con = self.get_initial_configuration()

    def get_initial_configuration(self):
        '''This returns the intial configuration of all
        models that this client can run. This information is used to
        build model runner instances'''
        r = call_restful_api('InitialConfiguration', base_url=self.url, key = self.key, method='GET')
        if isinstance(r, dict):
            return BioMAConfiguration.from_dict(r, client=self)
        else:
            logging.warning('Unable to retrieve Initial Configuration. Check you access level.')
            return {}
    
    def get_object_configuration(self, assembly_qualified_name:str):
        '''Returns all configurable items the passed BioMA object could possibly accept'''
        r = call_restful_api('DocumentObject', data= assembly_qualified_name, base_url=self.url, key = self.key, method='POST')
        if isinstance(r, dict):
            return r
        else:
            logging.warning('Unable to retrieve Object configuration. Check you access level.')
            return {}

    def get_object_subtypes(self, assembly_qualified_name:str):
        '''Returns all sub-types the passed BioMA object has on the server'''
        r = call_restful_api('DocumentType', data= assembly_qualified_name, base_url=self.url, key = self.key, method='POST')
        if isinstance(r, dict):
            return r
        else:
            #logging.warning('Unable to retrieve subtypes. Check you access level.')
            return {}

    def add_model_run(self, label:str, model:AbstractBioMAComponent):
        self.runs[label] = model


    def list_models(self):
        '''Returns a list of all ready-made models available on the server.
        
        each string is the assembly-qualified name of a model'''
        r = call_restful_api('ListModels', base_url=self.url, key = self.key, method='GET')
        if isinstance(r, list):
            return [BioMAConfiguration(i, self) for i in r]
        else:
            logging.warning('Unable to retrieve available models. Check you access level.')
            return []
        
    def run(self, idx):
        run = self.runs.get(idx)
        r = call_restful_api('ModelExecution', data= run.json_serialize(), base_url=self.url, key = self.key, method='POST')
        if isinstance(r, dict):
            return ModelResult.from_dict(r)
        else:
            return r

    def run_all(self):
        '''Runs all the loaded model runs'''
        # for each model run, we call the ModelExecution endpoint
        for idx, run in self.runs.items():
            r = call_restful_api('ModelExecution', data= run.json_serialize(), base_url=self.url, key = self.key, method='POST')
            if isinstance(r, dict):
                out = ModelResult.from_dict(r)
                self.outputs[idx] = out
            else:
                logging.warning(f'Model did not return a valid BioMA output for run #{idx}')
                self.outputs[idx] = out



class CalibrationClient():

    def __init__(self, url, key = None, env_variable_name = 'BIOMA_API_KEY'):
        self.url =url
        self.jobs = []
        if key is None:
            env_key = os.environ.get(env_variable_name)
            if env_key is None:
                logging.warning('No API key found in environment')
                self.key = None
            else:
                logging.info(f'Using variable {env_variable_name} from local environment as API key')
                self.key = env_key
        else:
            self.key = key

    def create_job(self, reference_data, alg_params):
        pass


import inspect
import logging
import pandas as pd
import json

from biomatools.utils.abs import AbstractBioMATable
from biomatools.utils.formats import (
    is_bioma_location,
    is_bioma_table,
    is_bioma_weather,
    prepare_df,
)
from biomatools.utils.naming import (
    BIOMA_LOCATION_COLUMNS,
    BIOMA_LOCATION_NAME,
    BIOMA_WEATHER_COLUMNS,
    BIOMA_WEATHER_NAME,
)


class InvalidWeatherDictionaryError(Exception):
    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super().__init__(message)


class InvalidLocationDictionaryError(Exception):
    def __init__(self, message):
        # Call the base class constructor with the parameters it needs
        super().__init__(message)


class BioMATable(AbstractBioMATable):
    """A Data Table in the BioMA application"""

    def __init__(self, Name: str, Rows, ColumnNames, NotNullFields = None):

        self.Name = Name
        self.Rows = Rows
        self.ColumnNames = ColumnNames
        self.NotNullFields = NotNullFields # this is apparently nullable

    @classmethod
    def from_dict(cls, env):
        """This is mostly used to build the object from the output returned by Web Services"""
        if is_bioma_table(env):
            return cls(
                **{
                    k: v
                    for k, v in env.items()
                    if k in inspect.signature(cls).parameters
                }
            )

    @classmethod
    def from_series(cls, name, **kwargs):
        """Builds the table concatenating multiple named series passed as argument"""
        cols = [label for label, _ in kwargs.items()]
        data = [value for _, value in kwargs.items()]
        df = pd.concat(data, axis=1).reset_index()
        df.columns = cols
        return cls.from_df(name, df)

    @classmethod
    def from_df(cls, Name: str, env: pd.DataFrame, include_index=False):
        """This is mostly used to build the object from the data included in your beloved Excel files"""
        columns = env.columns
        # here columns that have at least one value
        not_null_columns = [
            colname for colname in columns if any(pd.not_na(env[colname]))
        ]
        data = env.to_records(index=include_index)
        return cls(Name, Rows=data, ColunNames=columns, NotNullFields=not_null_columns)

    def to_df(self):
        return pd.DataFrame.from_records(self.Rows, columns=self.ColumnNames)

    def toJSON(self):
        return json.dumps(self, default=lambda o: vars(o), sort_keys=True)

    def json_serialize(self):
        """Tables have to be formatted as a JSON *string* to be
        digested by the BioMA web services, as weird as it sounds!"""
        return self.toJSON()


class BioMAWeatherTable(BioMATable):

    def __init__(self, Rows, NotNullFields=[]):
        super().__init__(
            BIOMA_WEATHER_NAME,
            Rows,
            ColumnNames=BIOMA_WEATHER_COLUMNS,
            NotNullFields=NotNullFields,
        )

    @classmethod
    def from_dict(cls, env):
        """This is mostly used to build the object from the output returned by Web Services"""
        if is_bioma_weather(env):
            return cls(
                **{
                    k: v
                    for k, v in env.items()
                    if k in inspect.signature(cls).parameters
                }
            )
        else:
            raise InvalidWeatherDictionaryError(
                "input dictionary is not a valid BioMA weather object"
            )

    @classmethod
    def from_df(cls, env, include_index=False, renaming=None):
        w = prepare_df(df=env, col_list=BIOMA_WEATHER_COLUMNS, renaming=renaming)
        return super().from_df(BIOMA_WEATHER_NAME, w, include_index)

    @classmethod
    def from_series(
        cls,
        AtmosphericPressure=[],
        CO2concentration=[],
        WaterTemperatureMinimum=[],
        WaterTemperatureMaximum=[],
        IsEvapotrPresent=[],
        IsVpdPresent=[],
        IsHumidityPresent=[],
        AirRelativeHumidityMinimum=[],
        Windspeed=[],
        VapourPressureDeficit=[],
        SnowDepth=[],
        AirTemperatureMinimum=[],
        AirTemperatureMaximum=[],
        SoilEvaporation=[],
        ReferenceEvapotranspiration=[],
        SurfaceOzoneConcentration=[],
        GlobalSolarRadiation=[],
        IsCO2Present=[],
        Date=[],
        Average_temperature=[],
        Precipitation=[],
        AirRelativeHumidityMaximum=[],
        LeafWetnessDuration=[],
        Grid_no=[],
    ):
        print(locals())
        return super().from_series(BIOMA_WEATHER_NAME, **locals())


class BioMALocationTable(BioMATable):

    def __init__(self, Rows, NotNullFields=[]):
        super().__init__(
            BIOMA_LOCATION_NAME,
            Rows,
            ColumnNames=BIOMA_LOCATION_COLUMNS,
            NotNullFields=NotNullFields,
        )

    @classmethod
    def from_dict(cls, env):
        """This is mostly used to build the object from the output returned by Web Services"""
        if is_bioma_location(env):
            return cls(
                **{
                    k: v
                    for k, v in env.items()
                    if k in inspect.signature(cls).parameters
                }
            )
        else:
            raise InvalidLocationDictionaryError(
                "input dictionary is not a valid BioMA location object"
            )

    @classmethod
    def from_df(cls, env, include_index=False, renaming=None):
        w = prepare_df(df=env, col_list=BIOMA_LOCATION_COLUMNS, renaming=renaming)
        return super().from_df(BIOMA_LOCATION_NAME, env, include_index)



from setuptools import setup

with open("README.md", 'r') as f:
    long_description = f.read()

setup(
   name='bioma-modelling',
   version='0.0.1',
   description='Python library to streamline interaction with the BioMA APIs, providing a pythonic facade to data and service access.',
   license="MIT",
   long_description=long_description,
   long_description_content_type="text/markdown",
   author='CREA',
   author_email='dario.denart@crea.gov.it',
   url="https://dev.azure.com/dariodenart/_git/BioMA%20Python%20Toolkit",
   packages=['biomatools', 'biomatools.utils'],
   install_requires=['pandas','requests', 'retry'], #external packages as dependencies
)


import inspect
import json
import logging

from biomatools.utils.abs import (
    AbstractBioMAComponent,
    AbstractBioMAClient,
    AbstractBioMAObject,
)
from biomatools.utils.formats import (
    check_bioma_type,
    is_bioma_table,
    json_serialize_component,
)
from biomatools.utils.naming import BIOMA_SIMPLE_TYPES


class BioMAConfiguration(AbstractBioMAComponent):
    """A Configuration item inside the BioMA call"""

    def __init__(
        self,
        assembly_qualified_name,
        client: AbstractBioMAClient = None,
        recursive: bool = False,
    ):
        # assembly qualified name
        self.BioMAtype = assembly_qualified_name
        # configuration variables
        self.vars = {}
        self.subtypes = []
        self.configurations = {}
        if client:
            self.load_subtypes(client, recursive)
            self.load_configurations(client)

    def load_configurations(self, client: AbstractBioMAClient):
        """
        Retrieves all possible configuration items of the current BioMA component
        supported by the passed client
        """
        if any(basetype in self.BioMAtype for basetype in BIOMA_SIMPLE_TYPES):
            self.configurations = {}
        else:
            self.configurations = client.get_object_configuration(self.BioMAtype)

    def load_subtypes(self, client: AbstractBioMAClient, recursive: bool = False):
        """
        Retrieves all possible subtypes of the current BioMA component
        supported by the passed client
        """
        if any(basetype in self.BioMAtype for basetype in BIOMA_SIMPLE_TYPES):
            self.subtypes = []
        elif recursive:
            self.subtypes = [
                BioMAConfiguration(i, client)
                for i in client.get_object_subtypes(self.BioMAtype)
            ]
        else:
            self.subtypes = [
                BioMAConfiguration(i)
                for i in client.get_object_subtypes(self.BioMAtype)
            ]

    def set_param(
        self, parname: str, parvalue=None, client: AbstractBioMAClient = None
    ):
        self.vars[parname] = parvalue
        if parname in self.configurations:
            # we now that parameter! Let's check its type
            if isinstance(parvalue, AbstractBioMAComponent) and (client is not None):
                if not check_bioma_type(
                    self.configurations.get(parname), parvalue, client
                ):
                    logging.warning(
                        f"Incompatible BioMA type at {parname}: expected {self.configurations.get(parname).BioMAtype}, found {parvalue.BioMAtype}"
                    )
        else:
            logging.warning(
                f"Parameter {parname} not found in {self.BioMAtype} configuration"
            )

    def describe(self):
        '''It returns a human understandable view of the component'''
        shorthand = self.BioMAtype.split(',')[0]
        out= f'Component {shorthand}\n\nFully qualified name: {self.BioMAtype}'
        if len(self.vars) > 0:
            out = out + '\n\nConfigured Variables\n'
            for k, v in self.vars.items():
                out = out + f'+ {k} : {v}\n'
            out = out + '\n'
        if len(self.configurations)>0:
            out = out + '\n\nAllowed Configurations\n'
            for k, v in self.configurations.items():
                out = out + f'+ {k} : {v}\n'
            out = out + '\n'
        if len(self.subtypes)>0:
            out = out + '\n\nSubtypes\n'
            for  v in self.subtypes:
                out = out + f'+ {v}\n'
            out = out + '\n'
        return out

    def toJSON(self):
        return json.dumps(
            {"@type": self.BioMAtype} | self.vars,
            default=json_serialize_component,
            sort_keys=True,
            indent=4,
        )

    def toDict(self):
        return {"@type": self.BioMAtype} | {self.vars}

    def json_serialize(self):
        serialized_vars = {}
        for k, v in self.vars.items():
            if isinstance(v, AbstractBioMAObject):
                # serialization is recursive!
                serialized_vars[k] = v.json_serialize()
            else:
                serialized_vars[k] = v

        return {"@type": self.BioMAtype} | serialized_vars

    @classmethod
    def from_dict(cls, env: dict, client:AbstractBioMAClient=None):
        """This is mostly used to build the object from the output returned by Web Services"""
        if isinstance(env, dict):
            if "@type" in env:
                out = cls(env["@type"], client=client)
                for parname, parvalue in env.items():
                    if parname != "@type":
                        if isinstance(parvalue, dict):
                            # is it another BioMA component?
                            if (
                                "@type" in parvalue
                            ):  # @type has already been taken care of
                                out.set_param(
                                    parname=parname,
                                    parvalue=BioMAConfiguration.from_dict(parvalue),
                                )
                            else:
                                out.set_param(parname=parname, parvalue=parvalue)
                        else:
                            # here it's anything but a dictionary
                            out.set_param(parname=parname, parvalue=parvalue)
                    # if parname is @type we don't do anything
                return out
            else:
                raise ValueError(f"No @type attribute in the provided dictionary")

        else:
            raise ValueError(f"Expected a dictionary, got {type(env)}")
        
    
    def __str__(self):
        return f'{type(self)} of type {self.BioMAtype.split(",")[0]}'
    def __repr__(self):
        return self.__str__()



import json
import requests
from urllib.parse import urljoin

def call_restful_api(function, data={}, params = {}, base_url=None, key=None, method='POST'):
    '''
    Calls a RESTful API
    
        Parameters:
            function (string): the name of the API method to invoke
            data (dictionary): the data payload for POST requests
            params (dictrionary): key/value pairs to be translated into http parameters
            bas_url (string): the base addres of the API
            key (string): the service key you registered
            method(string): the HTTP method you want to use, defaults to POST

        Returns:
            service response (str or dictionary): Web service response, either a dictionary or a string.
    
    '''
    function_url = urljoin(base_url, function)
    # set request headers    
    headers = {'Ocp-Apim-Subscription-Key': key}
    response = None
    # invoke the Web Service
    if method in ['POST', 'post']:
        headers['Content-Type'] =  'application/json'
        if isinstance(data, str) or isinstance(data, bytes):
            response = requests.post(function_url, data=str(data), params=params, headers=headers)
        else:
            # serialize request data into a json object if needed
            payload = json.dumps(data)
            response = requests.post(function_url, data=payload, params=params, headers=headers)
    else:
        response=requests.get(function_url, params=params, headers=headers)
    if (response.status_code == 200) or (response.status_code == 202): # the 20X status means 'success'
        try:
            return response.json() 
        except:
            try:
                return response.text
            except:
                print('Unable to decode server response')
    else:
        print('Request failed with code ' + str(response.status_code))
    return None